from typing import Any, Dict, List, Literal, Optional, Type, Union

from pydantic import AliasChoices, ConfigDict, Field

from inference.core.workflows.execution_engine.entities.base import (
    Batch,
    OutputDefinition,
    WorkflowImageData,
)
from inference.core.workflows.execution_engine.entities.types import (
    OBJECT_DETECTION_PREDICTION_KIND,
    ImageInputField,
    StepOutputImageSelector,
    WorkflowImageSelector,
    StepOutputSelector
)
from inference.core.workflows.prototypes.block import (
    BlockResult,
    WorkflowBlock,
    WorkflowBlockManifest,
)
from smart_reid.core import SmartTrack, SmartTrackConfig
from smart_reid.embed import Embedder

LONG_DESCRIPTION = """
Detect the location of barcodes in an image.

This block is useful for manufacturing and consumer packaged goods projects where you 
need to detect a barcode region in an image. You can then apply Crop block to isolate 
each barcode then apply further processing (i.e. OCR of the characters on a barcode).
"""


class BlockManifest(WorkflowBlockManifest):
    model_config = ConfigDict(
        json_schema_extra={
            "name": "Smart ReID",
            "short_description": "Track and ReID objects across frames.",
            "long_description": LONG_DESCRIPTION,
            "license": "Enterprise",
            "block_type": "tracker",
        }
    )
    type: Literal["SmartReID", "SmartReID"]
    images: Union[WorkflowImageSelector, StepOutputImageSelector] = ImageInputField
    predictions: StepOutputSelector(
        kind=[
           OBJECT_DETECTION_PREDICTION_KIND,
        ]
    ) = Field(
        title="Bounding Boxes",
        description="The output of a detection model describing the bounding boxes that will be used to track the objects.",
        examples=["$steps.my_object_detection_model.predictions"],
        validation_alias=AliasChoices("predictions", "detections"),
    )
    rematch_new_tracks: bool = Field(
        title="Rematch New Tracks",
        description="Whether to attempt to rematch new tracks to previous tracks.",
        default=True,
    )
    rematch_swapped_tracks: bool = Field(
        title="Rematch Swapped Tracks",
        description="Whether to attempt to rematch swapped tracks to previous tracks.",
        default=True,
    )
    track_activation_threshold: float = Field(
        title="Track Activation Threshold",
        description="The threshold for activating a track.",
        default=0.1,
    )
    lost_track_buffer: int = Field(
        title="Lost Track Buffer",
        description="The number of frames to wait before declaring a track lost.",
        default=1000,
    )
    minimum_matching_threshold: float = Field(
        title="Minimum Matching Threshold",
        description="The minimum threshold for matching tracks to detections.",
        default=0.99,
    )
    frame_rate: int = Field(
        title="Frame Rate",
        description="The frame rate of the video.",
        default=60,
    )
    minimum_consecutive_frames: int = Field(
        title="Minimum Consecutive Frames",
        description="The minimum number of consecutive frames a track must be visible for.",
        default=1,
    )



    @classmethod
    def accepts_batch_input(cls) -> bool:
        return True

    @classmethod
    def describe_outputs(cls) -> List[OutputDefinition]:
        return [
            OutputDefinition(
                name="tracks", kind=[OBJECT_DETECTION_PREDICTION_KIND]
            )
        ]

class SmartReIDBlock(WorkflowBlock):
    trackers = {}
    embedder = None
    def __init__(self):
        if self.embedder is None:
            self.embedder = Embedder()

    @classmethod
    def get_manifest(cls) -> Type[WorkflowBlockManifest]:
        return BlockManifest
    
    def run(
        self,
        images: Batch[WorkflowImageData],
        predictions: OBJECT_DETECTION_PREDICTION_KIND,
        rematch_new_tracks: bool,
        rematch_swapped_tracks: bool,
        track_activation_threshold: float,
        lost_track_buffer: int,
        minimum_matching_threshold: float,
        frame_rate: int,
        minimum_consecutive_frames: int,
    ) -> BlockResult:
        # apply configuration settings
        config = SmartTrackConfig(
            rematch_new_tracks=rematch_new_tracks,
            rematch_swapped_tracks=rematch_swapped_tracks,
            track_activation_threshold=track_activation_threshold,
            lost_track_buffer=10000,#lost_track_buffer,
            minimum_matching_threshold=minimum_matching_threshold,
            frame_rate=frame_rate,
            minimum_consecutive_frames=minimum_consecutive_frames,
        )
        inference_images = [i.to_inference_format(numpy_preferred=True).get("value") for i in images]
        stream_ids = [i.video_metadata.video_identifier for i in images]
        results = []
        for image, prediction, stream_id in zip(inference_images, predictions, stream_ids):
            tracker = self.trackers.get(stream_id, None)
            if tracker is None:
                tracker = SmartTrack(config, self.embedder)
                self.trackers[stream_id] = tracker
            tracker.configure(config)
            tracks = tracker.update_with_detections(prediction, image)
            results.append(tracks)
        return [{"tracks": tracks} for tracks in results]
    
def load_blocks() -> List[Type[WorkflowBlock]]:
    return [SmartReIDBlock]



from typing import List, Tuple

import numpy as np

from supervision.detection.core import Detections
from supervision.detection.utils import box_iou_batch
from smart_reid import matching
from smart_reid.basetrack import BaseTrack, TrackState
from smart_reid.kalman_filter import KalmanFilter
from smart_reid.strack import STrack
from smart_reid.embed import Embedder
from threading import Thread


def detections2boxes(detections: Detections) -> np.ndarray:
    """
    Convert Supervision Detections to numpy tensors for further computation.
    Args:
        detections (Detections): Detections/Targets in the format of sv.Detections.
    Returns:
        (np.ndarray): Detections as numpy tensors as in
            `(x_min, y_min, x_max, y_max, confidence, class_id)` order.
    """
    return np.hstack(
        (
            detections.xyxy,
            detections.confidence[:, np.newaxis],
            detections.class_id[:, np.newaxis],
        )
    )

class SmartTrackConfig:
    rematch_new_tracks: bool = True
    rematch_swapped_tracks: bool = True
    track_activation_threshold: float = 0.25
    lost_track_buffer: int = 30
    minimum_matching_threshold: float = 0.8
    frame_rate: int = 30
    minimum_consecutive_frames: int = 1
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        
class SmartTrack:
    """
    Initialize the ByteTrack object.

    <video controls>
        <source src="https://media.roboflow.com/supervision/video-examples/how-to/track-objects/annotate-video-with-traces.mp4" type="video/mp4">
    </video>

    Parameters:
        track_activation_threshold (float, optional): Detection confidence threshold
            for track activation. IncreasinRematched lostg track_activation_threshold improves accuracy
            and stability but might miss true detections. Decreasing it increases
            completeness but risks introducing noise and instability.
        lost_track_buffer (int, optional): Number of frames to buffer when a track is lost.
            Increasing lost_track_buffer enhances occlusion handling, significantly
            reducing the likelihood of track fragmentation or disappearance caused
            by brief detection gaps.
        minimum_matching_threshold (float, optional): Threshold for matching tracks with detections.
            Increasing minimum_matching_threshold improves accuracy but risks fragmentation.
            Decreasing it improves completeness but risks false positives and drift.
        frame_rate (int, optional): The frame rate of the video.
        minimum_consecutive_frames (int, optional): Number of consecutive frames that an object must
            be tracked before it is considered a 'valid' track.
            Increasing minimum_consecutive_frames prevents the creation of accidental tracks from
            false detection or double detection, but risks missing shorter tracks.
    """  # noqa: E501 // docs

    def __init__(
        self,
        config: SmartTrackConfig,
        embedder: Embedder
    ):
        self.configure(config)
        self.frame_id = 0
        self.kalman_filter = KalmanFilter()

        self.tracked_tracks: List[STrack] = []
        self.lost_tracks: List[STrack] = []
        self.removed_tracks: List[STrack] = []
        self.all_matched = True
        self.last_tracks = 0
        self.singleton_id = None
        self.embedder = embedder

    def configure(self, config: SmartTrackConfig):
        self.track_activation_threshold = config.track_activation_threshold
        self.lost_track_buffer = config.lost_track_buffer
        self.minimum_matching_threshold = config.minimum_matching_threshold
        self.frame_rate = config.frame_rate
        self.minimum_consecutive_frames = config.minimum_consecutive_frames
        self.max_time_lost = int(self.frame_rate / 30.0 * self.lost_track_buffer)
        self.det_thresh = self.track_activation_threshold + 0.1
        self.rematch_new_tracks = config.rematch_new_tracks
        self.rematch_swapped_tracks = config.rematch_swapped_tracks

    def update_with_detections(self, detections: Detections, frame) -> Detections:
        """
        Updates the tracker with the provided detections and returns the updated
        detection results.

        Args:
            detections (Detections): The detections to pass through the tracker.

        Example:
            ```python
            import supervision as sv
            from ultralytics import YOLO

            model = YOLO(<MODEL_PATH>)
            tracker = sv.ByteTrack()

            bounding_box_annotator = sv.BoundingBoxAnnotator()
            label_annotator = sv.LabelAnnotator()

            def callback(frame: np.ndarray, index: int) -> np.ndarray:
                results = model(frame)[0]
                detections = sv.Detections.from_ultralytics(results)
                detections = tracker.update_with_detections(detections)

                labels = [f"#{tracker_id}" for tracker_id in detections.tracker_id]

                annotated_frame = bounding_box_annotator.annotate(
                    scene=frame.copy(), detections=detections)
                annotated_frame = label_annotator.annotate(
                    scene=annotated_frame, detections=detections, labels=labels)
                return annotated_frame

            sv.process_video(
                source_path=<SOURCE_VIDEO_PATH>,
                target_path=<TARGET_VIDEO_PATH>,
                callback=callback
            )
            ```
        """

        tensors = detections2boxes(detections=detections)
        tracks = self.update_with_tensors(tensors=tensors, frame=frame)

        if len(tracks) > 0:
            detection_bounding_boxes = np.asarray([det[:4] for det in tensors])
            track_bounding_boxes = np.asarray([track.tlbr for track in tracks])

            ious = box_iou_batch(detection_bounding_boxes, track_bounding_boxes)

            iou_costs = 1 - ious

            matches, _, _ = matching.linear_assignment(iou_costs, 0.5)
            detections.tracker_id = np.full(len(detections), -1, dtype=int)
            for i_detection, i_track in matches:
                detections.tracker_id[i_detection] = int(
                    tracks[i_track].external_track_id
                )
            detections = detections[detections.tracker_id != -1]
            detections.tracker_id = detections.tracker_id[detections.tracker_id != -1]

            return detections

        else:
            detections = Detections.empty()
            detections.tracker_id = np.array([], dtype=int)
            detections.confidence = np.array([])
            detections.class_id = np.array([])
            detections["class_name"] = []

            return detections

    def reset(self):
        """
        Resets the internal state of the ByteTrack tracker.

        This method clears the tracking data, including tracked, lost,
        and removed tracks, as well as resetting the frame counter. It's
        particularly useful when processing multiple videos sequentially,
        ensuring the tracker starts with a clean state for each new video.
        """
        self.frame_id = 0
        self.tracked_tracks: List[STrack] = []
        self.lost_tracks: List[STrack] = []
        self.removed_tracks: List[STrack] = []
        BaseTrack.reset_counter()
        STrack.reset_external_counter()

    def update_with_tensors(self, tensors: np.ndarray, frame) -> List[STrack]:
        """
        Updates the tracker with the provided tensors and returns the updated tracks.

        Parameters:
            tensors: The new tensors to update with.

        Returns:
            List[STrack]: Updated tracks.
        """
        self.frame_id += 1
        # print(f"Updating with frame: {self.frame_id}")
        activated_starcks = []
        refind_stracks = []
        lost_stracks = []
        removed_stracks = []

        class_ids = tensors[:, 5]
        scores = tensors[:, 4]
        bboxes = tensors[:, :4]

        remain_inds = scores > self.track_activation_threshold
        inds_low = scores > 0.1
        inds_high = scores < self.track_activation_threshold

        inds_second = np.logical_and(inds_low, inds_high)
        dets_second = bboxes[inds_second]
        dets = bboxes[remain_inds]
        scores_keep = scores[remain_inds]
        scores_second = scores[inds_second]

        class_ids_keep = class_ids[remain_inds]
        class_ids_second = class_ids[inds_second]

        if len(dets) > 0:
            #print(dets[0])
            """Detections"""
            detections = [
                STrack(STrack.tlbr_to_tlwh(tlbr), s, c, self.minimum_consecutive_frames, frame[int(tlbr[1]):int(tlbr[3]), int(tlbr[0]):int(tlbr[2])])
                for (tlbr, s, c) in zip(dets, scores_keep, class_ids_keep)
            ]
        else:
            detections = []

        """ Add newly detected tracklets to tracked_stracks"""
        unconfirmed = []
        tracked_stracks = []  # type: list[STrack]

        for track in self.tracked_tracks:
            if not track.is_activated:
                unconfirmed.append(track)
            else:
                tracked_stracks.append(track)

        """ Step 2: First association, with high score detection boxes"""
        strack_pool = joint_tracks(tracked_stracks, self.lost_tracks)
        # Predict the current location with KF
        STrack.multi_predict(strack_pool)
        dists = matching.iou_distance(strack_pool, detections)

        dists = matching.fuse_score(dists, detections)
        matches, u_track, u_detection = matching.linear_assignment(
            dists, thresh=self.minimum_matching_threshold
        )

        for itracked, idet in matches:
            track = strack_pool[itracked]
            det = detections[idet]
            if track.state == TrackState.Tracked:
                track.update(detections[idet], self.frame_id, update_patch=len(self.tracked_tracks) > 1)
                # track.patch = [t for t in strack_pool if t.external_track_id == track.external_track_id][0].patch
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                # track.patch = [t for t in strack_pool if t.external_track_id == track.external_track_id][0].patch
                refind_stracks.append(track)
            
        """ Step 3: Second association, with low score detection boxes"""
        # association the untrack to the low score detections
        if len(dets_second) > 0:
            """Detections"""
            detections_second = [
                STrack(STrack.tlbr_to_tlwh(tlbr), s, c, self.minimum_consecutive_frames, frame[int(tlbr[1]):int(tlbr[3]), int(tlbr[0]):int(tlbr[2])])
                for (tlbr, s, c) in zip(dets_second, scores_second, class_ids_second)
            ]
        else:
            detections_second = []
        r_tracked_stracks = [
            strack_pool[i]
            for i in u_track
            if strack_pool[i].state == TrackState.Tracked
        ]
        dists = matching.iou_distance(r_tracked_stracks, detections_second)

        matches, u_track, u_detection_second = matching.linear_assignment(
            dists, thresh=0.5
        )
        for itracked, idet in matches:
            track = r_tracked_stracks[itracked]
            det = detections_second[idet]
            if track.state == TrackState.Tracked:
                track.update(det, self.frame_id, update_patch=len(self.tracked_tracks) > 1)
                # track.patch = [t for t in strack_pool if t.external_track_id == track.external_track_id][0].patch
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                # track.patch = [t for t in strack_pool if t.external_track_id == track.external_track_id][0].patch
                refind_stracks.append(track)

        for it in u_track:
            track = r_tracked_stracks[it]
            if not track.state == TrackState.Lost:
                track.mark_lost()
                lost_stracks.append(track)

        """Deal with unconfirmed tracks, usually tracks with only one beginning frame"""
        detections = [detections[i] for i in u_detection]
        if len(detections) > 0:
            '''print(f"Unconfirmed: {len(detections)}")'''
        dists = matching.iou_distance(unconfirmed, detections)

        dists = matching.fuse_score(dists, detections)
        
        matches, u_unconfirmed, u_detection = matching.linear_assignment(
            dists, thresh=0.7
        )
        for itracked, idet in matches:
            unconfirmed[itracked].update(detections[idet], self.frame_id, update_patch=len(self.tracked_tracks) > 1)
            activated_starcks.append(unconfirmed[itracked])
        for it in u_unconfirmed:
            track = unconfirmed[it]
            track.mark_removed()
            removed_stracks.append(track)

        """ Step 4: Init new stracks"""
        # embed tracks which have not been matched with any detections
        new_stracks = []
        for i in u_detection:
            track = detections[i]
            if track.score < self.det_thresh:
                continue
            track.activate(self.kalman_filter, self.frame_id)
            activated_starcks.append(track)  
            new_stracks.append(track)

        if len(u_detection) == 1 and len(self.lost_tracks) >= 1:
            best_lost_track = min(self.lost_tracks, key=lambda x: x.external_track_id)
            i_id = new_stracks[0].internal_track_id
            e_id = new_stracks[0].external_track_id
            new_stracks[0].external_track_id = best_lost_track.external_track_id
            new_stracks[0].internal_track_id = best_lost_track.internal_track_id
            best_lost_track.external_track_id = e_id
            best_lost_track.internal_track_id = i_id


            

        """ Step 5: Update state"""
        for track in self.lost_tracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)
                # print(f"Removed track {track.external_track_id} due to time lost")

        

        # run sam reid if one detection shares significant iou with multiple tracks
        if len(lost_stracks) > 0:
            self.all_matched = False
        if self.rematch_swapped_tracks and len(lost_stracks) == 0 and len(self.lost_tracks) > 0 and len(self.tracked_tracks) > 0 and not self.all_matched:
            # print("Rematching swapped tracks")
            self.all_matched = True
            track_pool = joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks)
            track_dists = matching.iou_distance(track_pool, self.tracked_tracks)
            dt = track_dists
            dets_to_rematch = []
            tracks_to_rematch = []
            for idet in range(dt.shape[0]):
                to_match = []
                for itrack in range(dt.shape[1]):
                    if dt[idet, itrack] < self.minimum_matching_threshold:
                        to_match.append(itrack)
                    if len(to_match) > 1:
                        dets_to_rematch.append(track_pool[idet])
                        tracks_to_rematch.extend([self.tracked_tracks[tm] for tm in to_match])

            rematch_thread = Thread(target=self.embedder.async_rematch, args=(tracks_to_rematch, dets_to_rematch, frame, lambda x: self.rematch_tracks_swapped(x), 0.4))
            rematch_thread.start()

        self.tracked_tracks = [
            t for t in self.tracked_tracks if t.state == TrackState.Tracked
        ]
        self.tracked_tracks = joint_tracks(self.tracked_tracks, new_stracks)
        self.tracked_tracks = joint_tracks(self.tracked_tracks, activated_starcks)
        self.tracked_tracks = joint_tracks(self.tracked_tracks, refind_stracks)

        # prevent from changing single tracked tack's id while only one track is present
        # if len(self.tracked_tracks) > 1:
        #     self.singleton_id = None
        # if len(self.tracked_tracks) == 1 and self.singleton_id is None:
        #     self.singleton_id = self.tracked_tracks[0].external_track_id
        # if len(self.tracked_tracks) == 1 and self.singleton_id is not None:
        #     self.tracked_tracks[0].external_track_id = self.singleton_id

        # update patches of tracked tracks if there are two tracks
        # if len(self.tracked_tracks) > 1:
        #     for i in range(len(self.tracked_tracks)):
        #         for j in range(i+1, len(self.tracked_tracks)):
        #             self.tracked_tracks[i].patch = self.tracked_tracks[j].new_patch

        # if self.rematch_new_tracks and len(self.tracked_tracks) > self.last_tracks:
        #     print("Rematching new tracks")
        #     rematch_thread = Thread(target=async_rematch, args=(self.lost_tracks, self.tracked_tracks, frame, lambda x: self.rematch_tracks_lost(x), 0.4))
        #     rematch_thread.start()
        self.lost_tracks.extend(lost_stracks)
        self.lost_tracks = sub_tracks(self.lost_tracks, self.tracked_tracks)
        self.lost_tracks = sub_tracks(self.lost_tracks, self.removed_tracks)
        self.removed_tracks = removed_stracks
        self.last_tracks = len(self.tracked_tracks)

        if len(self.tracked_tracks) > 1 and self.rematch_new_tracks and len(new_stracks) > 0:
            # asynchronously rematch the tracks
            # print("Rematching new tracks")
            rematch_thread = Thread(target=self.embedder.async_rematch, args=([track for track in self.tracked_tracks if track not in new_stracks],  self.tracked_tracks, frame, lambda x: self.rematch_tracks_lost(x)))
            rematch_thread.start()

        ids = [t.external_track_id for t in self.tracked_tracks]
        # print("IDS", ids)
        ids_not_included = [i for i in [1, 2] if i not in ids]
        total_tracks = [t for t in joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks) if t.external_track_id not in [1,2]]
        #  print("ID NOT INCLUDED", ids_not_included)
        if len(ids_not_included) > 0 and len(self.tracked_tracks) > 1:
            # find the track with the lowest external id
            # print("TOTAL TRACKS", len(total_tracks))
            for i in ids_not_included:
                avail_tracks = [t for t in self.tracked_tracks if t.external_track_id not in [1,2]]
                if len(avail_tracks) == 0:
                    continue
                min_track = min(avail_tracks, key=lambda x: x.external_track_id)
                min_track.activate(self.kalman_filter, self.frame_id)
                min_track.external_track_id = i
                # print("MIN TRACK", min_track.external_track_id)
                self.tracked_tracks = joint_tracks(self.tracked_tracks, [min_track])
                # print("TRACKED TRACKS", len(self.tracked_tracks))
                total_tracks = sub_tracks(total_tracks, [min_track])
                lost_id_tracks = [t for t in self.lost_tracks if t.external_track_id == i]
                lost_id_track = lost_id_tracks[0] if lost_id_tracks else None
                if lost_id_track:
                    self.lost_tracks = sub_tracks(self.lost_tracks, [lost_id_track])
                    total_tracks = sub_tracks(total_tracks, [lost_id_track])
                removed_id_tracks = [t for t in self.removed_tracks if t.external_track_id == i]
                removed_id_track = removed_id_tracks[0] if removed_id_tracks else None
                if removed_id_track:
                    self.removed_tracks = sub_tracks(self.removed_tracks, [removed_id_track])
                    total_tracks = sub_tracks(total_tracks, [removed_id_track])
                # print("TOTAL TRACKS", len(total_tracks))

        output_stracks = [track for track in self.tracked_tracks if track.is_activated]
        # for track in output_stracks:
            # print(f"Track {track.external_track_id} has {track.continuous_track_count} continuous track count")
        # print(f"Tracked: {len(self.tracked_tracks)} Lost: {len(self.lost_tracks)} Removed: {len(self.removed_tracks)} New: {len(new_stracks)}")

        return output_stracks
    
    def rematch_tracks_lost(self, matches):
        matches = [(i, j) for i, j in matches if i != j]
        for i, j in matches:
            i_tracks = [t for t in joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks) if t.external_track_id == i]
            if len(i_tracks) == 0:
                continue
            i_track = i_tracks[0]
            j_tracks = [t for t in joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks) if t.external_track_id == j]
            if len(j_tracks) == 0:
                continue
            j_track = j_tracks[0]
            j_track.external_track_id = i
            j_internal = j_track.internal_track_id
            j_patch = j_track.patch
            j_state = j_track.state
            j_track.internal_track_id = i_track.internal_track_id
            j_track.state = i_track.state
            j_track.patch = i_track.patch
            j_track.continuous_track_count = 0
            i_track.external_track_id = j
            i_track.patch = j_patch
            i_track.state = j_state
            i_track.internal_track_id = j_internal
            i_track.continuous_track_count = 0
            self.tracked_tracks = joint_tracks(self.tracked_tracks, [j_track])
            self.tracked_tracks = sub_tracks(self.tracked_tracks, [i_track])
            if i_track.state == TrackState.Lost:
                self.lost_tracks = sub_tracks(self.lost_tracks, [i_track])
            elif j_track.state == TrackState.Lost:
                self.lost_tracks = sub_tracks(self.lost_tracks, [j_track])
            else:
                self.tracked_tracks = joint_tracks(self.tracked_tracks, [i_track])



    def rematch_tracks_swapped(self, matches):
        matches = [(i, j) for i, j in matches if i != j]
        for x, m in enumerate(matches):
            i, j = m
            i_in = False
            j_in = False
            for t in self.tracked_tracks:
                if j == t.external_track_id:
                    j_in = True
                if i == t.external_track_id:
                    i_in = True
            if i_in and j_in and (j, i) not in matches:
                matches.remove(m)

        for i, j in matches:
            i_tracks = [t for t in joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks) if t.external_track_id == i]
            if len(i_tracks) == 0:
                continue
            i_track = i_tracks[0]
            j_tracks = [t for t in joint_tracks(joint_tracks(self.tracked_tracks, self.lost_tracks), self.removed_tracks) if t.external_track_id == j]
            if len(j_tracks) == 0:
                continue
            j_track = j_tracks[0]
            j_track.external_track_id = i
            j_internal = j_track.internal_track_id
            j_patch = j_track.patch
            j_state = j_track.state
            j_track.internal_track_id = i_track.internal_track_id
            j_track.state = i_track.state
            j_track.patch = i_track.patch
            j_track.continuous_track_count = 0
            i_track.external_track_id = j
            i_track.patch = j_patch
            i_track.state = j_state
            i_track.internal_track_id = j_internal
            i_track.continuous_track_count = 0


def joint_tracks(
    track_list_a: List[STrack], track_list_b: List[STrack]
) -> List[STrack]:
    """
    Joins two lists of tracks, ensuring that the resulting list does not
    contain tracks with duplicate internal_track_id values.

    Parameters:
        track_list_a: First list of tracks (with internal_track_id attribute).
        track_list_b: Second list of tracks (with internal_track_id attribute).

    Returns:
        Combined list of tracks from track_list_a and track_list_b
            without duplicate internal_track_id values.
    """
    seen_track_ids = set()
    result = []

    for track in track_list_a + track_list_b:
        if track.internal_track_id not in seen_track_ids:
            seen_track_ids.add(track.internal_track_id)
            result.append(track)

    return result


def sub_tracks(track_list_a: List, track_list_b: List) -> List[int]:
    """
    Returns a list of tracks from track_list_a after removing any tracks
    that share the same internal_track_id with tracks in track_list_b.

    Parameters:
        track_list_a: List of tracks (with internal_track_id attribute).
        track_list_b: List of tracks (with internal_track_id attribute) to
            be subtracted from track_list_a.
    Returns:
        List of remaining tracks from track_list_a after subtraction.
    """
    tracks = {track.internal_track_id: track for track in track_list_a}
    track_ids_b = {track.internal_track_id for track in track_list_b}

    for track_id in track_ids_b:
        tracks.pop(track_id, None)

    return list(tracks.values())


def remove_duplicate_tracks(tracks_a: List, tracks_b: List) -> Tuple[List, List]:
    pairwise_distance = matching.iou_distance(tracks_a, tracks_b)
    matching_pairs = np.where(pairwise_distance < 0.15)

    duplicates_a, duplicates_b = set(), set()
    for track_index_a, track_index_b in zip(*matching_pairs):
        time_a = tracks_a[track_index_a].frame_id - tracks_a[track_index_a].start_frame
        time_b = tracks_b[track_index_b].frame_id - tracks_b[track_index_b].start_frame
        if time_a > time_b:
            duplicates_b.add(track_index_b)
        else:
            duplicates_a.add(track_index_a)

    result_a = [
        track for index, track in enumerate(tracks_a) if index not in duplicates_a
    ]
    result_b = [
        track for index, track in enumerate(tracks_b) if index not in duplicates_b
    ]

    return result_a, result_b



# from segment_anything import sam_model_registry
from PIL import Image
from torchvision import transforms
import torch
import numpy as np
import cv2
from scipy.optimize import linear_sum_assignment
import open_clip
from pathlib import Path
from datetime import datetime
import mediapipe as mp
from image_similarity_measures.quality_metrics import rmse
import io
from super_image import PanModel, ImageLoader
import tempfile

class Embedder:
    def __init__(self):
        # Load the SAM model
        # model_type = "vit_h"
        # sam = sam_model_registry[model_type](checkpoint=os.path.join(os.path.dirname(__file__), "sam_vit_h_4b8939.pth"))
        # sam.to(device="cuda" if torch.cuda.is_available() else "cpu")

        # Extract the image encoder
        # image_encoder = sam.image_encoder
        self.smp_selfie_segmentation = mp.solutions.selfie_segmentation
        self.segment = self.smp_selfie_segmentation.SelfieSegmentation(model_selection = 0)
        self.model = PanModel.from_pretrained('eugenesiow/pan-bam', scale=3)


    def segment_subject(self, img):
        results = self.segment.process(img)
        mask = results.segmentation_mask
        threshold = 0.25
        binary_mask = mask > threshold
        if np.sum(binary_mask) != 0:
            img = img*binary_mask[0:img.shape[0],0:img.shape[1],np.newaxis]
        # cv2.imwrite(f"segmented+{datetime.now()}.png", img)
        return img

    # Load the MobileCLIP model
    # clip_model, _, clip_preprocess = open_clip.create_model_and_transforms('MobileCLIP-B', pretrained="datacompdr_lt")

    def match_tracks(self, cost_matrix, threshold=0.4):
        # Replace costs above the threshold with a large number to make them undesirable
        modified_cost_matrix = np.where(cost_matrix < threshold, cost_matrix, 1e9)
        
        # Use the Hungarian algorithm to find the optimal assignment
        row_indices, col_indices = linear_sum_assignment(modified_cost_matrix)
        
        # Filter out assignments that have a cost above the threshold
        matches = []
        for row, col in zip(row_indices, col_indices):
            if cost_matrix[row, col] < threshold:
                matches.append((row, col))
        
        return matches

    def async_rematch(self, oldtracks, newtracks, frame, callback, threshold=0.5):
        # embed newtracks and oldtracks
        # old_embeds = clip_embed([track for track in oldtracks], new=False)
        # new_embeds = clip_embed([track for track in newtracks], new=True)

        # perform matching between oldtracks and newtracks
        # cost_matrix = np.zeros((len(old_embeds), len(new_embeds)))
        # for i, old in enumerate(old_embeds):
        #     for j, new in enumerate(new_embeds):
        #         cost_matrix[i, j] = 1 - abs(np.dot(old, new) / (np.linalg.norm(old) * np.linalg.norm(new)))

        # print("Rematching: ", [track.external_track_id for track in oldtracks], [track.external_track_id for track in newtracks])
        cost_matrix = self.image_similarity_matrix(oldtracks, newtracks)
        # print("cost_matrix: ", cost_matrix)

        matches = self.match_tracks(cost_matrix, threshold)
        resmatches = [(oldtracks[i].external_track_id, newtracks[j].external_track_id) for i, j in matches]
        # print("Matches: ", resmatches)

        # change newtracks id to matching oldtrack id; return newtracks that matched

        callback(resmatches)

    def tracks_to_patches(self, tracks, new=False):
        imgs = []
        for track in tracks:
            patch = track.patch if not new else track.new_patch
            patch = self.model(ImageLoader.load_image(Image.fromarray(cv2.cvtColor(patch, cv2.COLOR_BGR2RGB))))#[0].detach().cpu().numpy().transpose(1, 2, 0).astype(np.uint8)
            patch = ImageLoader._process_image_to_save(patch)
            patch = cv2.convertScaleAbs(patch)
            img = self.segment_subject(patch)
            imgs.append(img)
        return imgs

    def image_similarity_matrix(self, tracks_a, tracks_b):
        imgs_a = self.tracks_to_patches(tracks_a, new=False)
        imgs_b = self.tracks_to_patches(tracks_b, new=True)
        if len(imgs_a) == 0 or len(imgs_b) == 0:
            return np.zeros((len(imgs_a), len(imgs_b)))
        H = max([img.shape[0] for img in [*imgs_a, *imgs_b]])
        W = max([img.shape[1] for img in [*imgs_a, *imgs_b]])
        # pad images to be the same size
        imgs_a = [cv2.copyMakeBorder(img,0,H - img.shape[0],0,W - img.shape[1],cv2.BORDER_CONSTANT,value=[0,0,0]) for img in imgs_a]
        imgs_b = [cv2.copyMakeBorder(img,0,H - img.shape[0],0,W - img.shape[1],cv2.BORDER_CONSTANT,value=[0,0,0]) for img in imgs_b]
        sim_matrix = np.zeros((len(imgs_a), len(imgs_b)))
        with tempfile.TemporaryDirectory() as tmpdirname:
            for i in range(len(imgs_a)):
                for j in range(len(imgs_b)):
                    image_similarity = rmse(imgs_a[i], imgs_b[j])
                    mask_iou = self.binaryMaskIOU(imgs_a[i], imgs_b[j])
                    sim_matrix[i, j] = image_similarity / (mask_iou if mask_iou != 0 else 1)
        return sim_matrix

    def binaryMaskIOU(self, mask1, mask2):   # From the question.
        mask1_area = np.count_nonzero(mask1 > 0)
        mask2_area = np.count_nonzero(mask2 > 0)
        if mask1_area == 0 or mask2_area == 0:
            return 0
        intersection = np.count_nonzero(np.logical_and( mask1>0,  mask2>0 ))
        iou = intersection/(mask1_area+mask2_area-intersection)
        return iou

    def clip_embed(self, predictions, new=False):
        res = []
        patches = []
        for track in predictions:
            patch = track.patch if not new else track.new_patch
            img = self.segment_subject(patch)
            img = Image.fromarray(img).convert('RGB')
            patches.append(self.clip_preprocess(img))
        if patches == []:
            return []
        patches_tensor = torch.stack(patches)
        with torch.no_grad(), torch.cuda.amp.autocast():
            embeddings = self.clip_model.encode_image(patches_tensor)
        for i, embedding in enumerate(embeddings):
            track = predictions[i]
            track.embedding = embedding
            res.append(embedding)

        return res

    def upscale_patches(self, patches):
        out = []
        for patch in patches:
            out.append(cv2.pyrUp(patch))
        return out



    # pred: [(x1, y1), (x2, y2), ..., (xn, yn)] for n keypoints
    # Keypoints order in COCO 2017 Keypoint Challenge annotation format:
    # 1. Nose
    # 2. Left Eye
    # 3. Right Eye
    # 4. Left Ear
    # 5. Right Ear
    # 6. Left Shoulder
    # 7. Right Shoulder
    # 8. Left Elbow
    # 9. Right Elbow
    # 10. Left Wrist
    # 11. Right Wrist
    # 12. Left Hip
    # 13. Right Hip
    # 14. Left Knee
    # 15. Right Knee
    # 16. Left Ankle
    # 17. Right Ankle
    def pose_embed(self, preds):
        # norm key points by finding dist from center of hips
        embeddings = []
        for pred in preds:
            hx = (pred[12][0] + pred[13][0]) / 2
            hy = (pred[12][1] + pred[13][1]) / 2
            normed = [(x - hx, y - hy) for x, y in pred]
            embedding = np.array([x for x, y in normed])#np.array(normed).flatten()
            embeddings.append(embedding)
        return embeddings

    def pose_cost(self, atracks, btracks):
        a_embeds = self.pose_embed([a.pose for a in atracks])
        b_embeds = self.pose_embed([b.pose for b in btracks])

        cost_matrix = np.zeros((len(a_embeds), len(b_embeds)))
        for i, a in enumerate(a_embeds):
            for j, b in enumerate(b_embeds):
                if np.sum(b) == 0 or np.sum(a) == 0:
                    cost_matrix[i, j] = 1
                else:
                    cost_matrix[i, j] = 1 - abs(np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b)))

        return cost_matrix


if __name__=="__main__": 
    # img = cv2.imread("test_balls.jpg")
    # img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    # print(img.shape)
    # embeddings = sam_embed(img, [[0, 0, 533, 400], [533, 0, 1066, 400], [0, 400, 533, 800]])
    # print(embeddings)
    # similarity = np.dot(embeddings[0], embeddings[1]) / (np.linalg.norm(embeddings[0]) * np.linalg.norm(embeddings[1]))
    # print(similarity)
    # similarity = np.dot(embeddings[0], embeddings[2]) / (np.linalg.norm(embeddings[0]) * np.linalg.norm(embeddings[2]))
    # print(similarity)

    v1 = [0.5, 1, -0.5, 1, 1, 0.5, -1, 0.5, 1, 0, -1, 0, 0.5, 0, -0.5, 0, 1, -1, -1, -1, 1, -2, -1, -2]
    v1 = [x for i,x in enumerate(v1) if i % 2 == 0]
    v2 = [0, 1, 0.5, 1, -0.5, 0.5, 1, 0.5, -1, 0.5, 1, 0, -0.5, 0, 0.5, 0, 0.5, -1, 0, -1, 0.5, -2, -1, -2]
    v2 = [x for i,x in enumerate(v2) if i % 2 == 0]
    sim = np.dot(np.array(v1), np.array(v2))/(np.linalg.norm(np.array(v1)) * np.linalg.norm(np.array(v2)))
    # print(sim)
    


from typing import List, Tuple

import numpy as np
from scipy.optimize import linear_sum_assignment

from supervision.detection.utils import box_iou_batch


def indices_to_matches(
    cost_matrix: np.ndarray, indices: np.ndarray, thresh: float
) -> Tuple[np.ndarray, tuple, tuple]:
    matched_cost = cost_matrix[tuple(zip(*indices))]
    matched_mask = matched_cost <= thresh

    matches = indices[matched_mask]
    unmatched_a = tuple(set(range(cost_matrix.shape[0])) - set(matches[:, 0]))
    unmatched_b = tuple(set(range(cost_matrix.shape[1])) - set(matches[:, 1]))
    return matches, unmatched_a, unmatched_b


def linear_assignment(
    cost_matrix: np.ndarray, thresh: float
) -> [np.ndarray, Tuple[int], Tuple[int, int]]:
    if cost_matrix.size == 0:
        return (
            np.empty((0, 2), dtype=int),
            tuple(range(cost_matrix.shape[0])),
            tuple(range(cost_matrix.shape[1])),
        )

    cost_matrix[cost_matrix > thresh] = thresh + 1e-4
    row_ind, col_ind = linear_sum_assignment(cost_matrix)
    indices = np.column_stack((row_ind, col_ind))

    return indices_to_matches(cost_matrix, indices, thresh)


"""
np.ndarray: Pairwise IoU of boxes from `boxes_true` and `boxes_detection`.
            `shape = (N, M)` where `N` is number of true objects and
            `M` is number of detected objects.
"""
def iou_distance(atracks: List, btracks: List) -> np.ndarray:
    if (len(atracks) > 0 and isinstance(atracks[0], np.ndarray)) or (
        len(btracks) > 0 and isinstance(btracks[0], np.ndarray)
    ):
        atlbrs = atracks
        btlbrs = btracks
    else:
        atlbrs = [track.tlbr for track in atracks]
        btlbrs = [track.tlbr for track in btracks]

    _ious = np.zeros((len(atlbrs), len(btlbrs)), dtype=np.float32)
    if _ious.size != 0:
        _ious = box_iou_batch(np.asarray(atlbrs), np.asarray(btlbrs))
    cost_matrix = 1 - _ious

    return cost_matrix


def fuse_score(cost_matrix: np.ndarray, detections: List) -> np.ndarray:
    if cost_matrix.size == 0:
        return cost_matrix
    iou_sim = 1 - cost_matrix
    det_scores = np.array([det.score for det in detections])
    det_scores = np.expand_dims(det_scores, axis=0).repeat(cost_matrix.shape[0], axis=0)
    fuse_sim = iou_sim #* det_scores
    fuse_cost = 1 - fuse_sim
    return fuse_cost



import numpy as np
from enum import Enum
from smart_reid.kalman_filter import KalmanFilter
from smart_reid.basetrack import BaseTrack, TrackState

class STrack(BaseTrack):
    shared_kalman = KalmanFilter()
    _external_count = 0

    def __init__(self, tlwh, score, class_ids, minimum_consecutive_frames, patch, embedding=None):
        # wait activate
        self._tlwh = np.asarray(tlwh, dtype=np.float32)
        self.kalman_filter = None
        self.mean, self.covariance = None, None
        self.is_activated = False

        self.score = score
        self.class_ids = class_ids
        self.tracklet_len = 0
        self.patch = patch

        self.external_track_id = -1

        self.minimum_consecutive_frames = minimum_consecutive_frames
        self.embedding = None
        self.new_patch = patch
        self.continuous_track_count = 0

    def predict(self):
        mean_state = self.mean.copy()
        if self.state != TrackState.Tracked:
            mean_state[7] = 0
        self.mean, self.covariance = self.kalman_filter.predict(
            mean_state, self.covariance
        )

    @staticmethod
    def multi_predict(stracks):
        if len(stracks) > 0:
            multi_mean = []
            multi_covariance = []
            for i, st in enumerate(stracks):
                multi_mean.append(st.mean.copy())
                multi_covariance.append(st.covariance)
                if st.state != TrackState.Tracked:
                    multi_mean[i][7] = 0

            multi_mean, multi_covariance = STrack.shared_kalman.multi_predict(
                np.asarray(multi_mean), np.asarray(multi_covariance)
            )
            for i, (mean, cov) in enumerate(zip(multi_mean, multi_covariance)):
                stracks[i].mean = mean
                stracks[i].covariance = cov

    def activate(self, kalman_filter, frame_id):
        """Start a new tracklet"""
        self.kalman_filter = kalman_filter
        self.internal_track_id = self.next_id()
        self.mean, self.covariance = self.kalman_filter.initiate(
            self.tlwh_to_xyah(self._tlwh)
        )

        self.tracklet_len = 0
        self.state = TrackState.Tracked
        if frame_id == 1:
            self.is_activated = True

        if self.minimum_consecutive_frames == 1:
            self.external_track_id = self.next_external_id()

        self.frame_id = frame_id
        self.start_frame = frame_id

    def re_activate(self, new_track, frame_id, new_id=False):
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_track.tlwh)
        )
        self.tracklet_len = 0
        self.state = TrackState.Tracked

        self.frame_id = frame_id
        if new_id:
            self.internal_track_id = self.next_id()
        self.score = new_track.score
        self.new_patch = new_track.patch

    def update(self, new_track, frame_id, update_patch=True):
        """
        Update a matched track
        :type new_track: STrack
        :type frame_id: int
        :type update_feature: bool
        :return:
        """
        self.frame_id = frame_id
        self.tracklet_len += 1

        new_tlwh = new_track.tlwh
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_tlwh)
        )
        self.state = TrackState.Tracked
        if self.tracklet_len == self.minimum_consecutive_frames:
            self.is_activated = True
            if self.external_track_id == -1:
                self.external_track_id = self.next_external_id()

        self.score = new_track.score
        self.new_patch = new_track.patch
        self.continuous_track_count += 1
        if self.continuous_track_count % 100 == 0 and update_patch:
            self.patch = new_track.patch

    @property
    def tlwh(self):
        """Get current position in bounding box format `(top left x, top left y,
        width, height)`.
        """
        if self.mean is None:
            return self._tlwh.copy()
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    @property
    def tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    @staticmethod
    def tlwh_to_xyah(tlwh):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = np.asarray(tlwh).copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret

    def to_xyah(self):
        return self.tlwh_to_xyah(self.tlwh)

    @staticmethod
    def next_external_id():
        STrack._external_count += 1
        return STrack._external_count

    @staticmethod
    def reset_external_counter():
        STrack._external_count = 0

    @staticmethod
    def tlbr_to_tlwh(tlbr):
        ret = np.asarray(tlbr).copy()
        ret[2:] -= ret[:2]
        return ret

    @staticmethod
    def tlwh_to_tlbr(tlwh):
        ret = np.asarray(tlwh).copy()
        ret[2:] += ret[:2]
        return ret

    def __repr__(self):
        return "OT_{}_({}-{})".format(
            self.internal_track_id, self.start_frame, self.end_frame
        )


import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


def read_requirements(path):
    if not isinstance(path, list):
        path = [path]
    requirements = []
    for p in path:
        with open(p) as fh:
            requirements.extend([line.strip() for line in fh])
    return requirements


setuptools.setup(
    name="smart_reid",
    version="0.1.7",
    author="Roboflow",
    author_email="help@roboflow.com",
    description="With no prior knowledge of machine learning or device-specific deployment, you can deploy a computer vision model to a range of devices and environments using Roboflow Inference CLI.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/roboflow/inference",
    packages=find_packages(
        where="./",
        exclude=(
            "examples",
        ),
    ),

    install_requires=read_requirements([
        "requirements.txt",
    ]),
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Typing :: Typed",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8,<=3.12",
)